Module source

Module source 

Structs

Env
Environment variable based Source.
File
File-backed Source that reads configuration from a file on disk.
RawData
Allows passing a raw string of data as a Source

Traits

A source of configuration data.