Env

Struct Env 

pub struct Env<Fmt> { /* private fields */ }
Environment variable based Source.

Env looks for variables starting with prefix and splits the remainder of the variable name by separator to produce nested keys in the resulting JSON object.

Implementations

Source

impl<Fmt> Env<Fmt>
where Fmt: Format,

pub fn new( prefix: impl Into<String>, separator: impl Into<Cow<'static, str>>, ) -> Self