pub struct Env<Fmt> { /* private fields */ }Expand description
Environment variable based Source.
Env looks for variables starting with prefix and splits the remainder
of the variable name by separator to produce nested keys in the
resulting JSON object.
Implementations